#include <stdio.h>
#include <avr/io.h>
#include "mmc.h"
#include "fat16.h"


//________________________________________________________________________________________________________________________________________
// Module name:			fat16.c 
// Compiler used:		avr-gcc 3.4.5
// Last Modifikation:	24.07.2007
// Version:				1.23
// Authors:				Stephan Busker 	
// Description:			Source files for FAT16 implementation with read and write-access using AVR-Mikrocontrollers
//						Copyright (C) 2007 Stephan Busker
//........................................................................................................................................
// Functions:			extern unsigned char 	InitFat16(void);		
//						unsigned char 			fopen_(unsigned char *fname,char mode, File *file);
//						void 					fclose_(File *file);
//						unsigned long 			fread_(void *buffer, unsigned long size, unsigned long count, File *file); 
//						unsigned long 			fwrite_(void *buffer, unsigned long size, unsigned long count, File *file);
//						int  					fseek_(File *file, long offset, int origin);
//	         			char 					fgetchar_(File *file);
//						unsigned char 			fputchar_(File *file,char c);
//						unsigned char 			fputs_(File *file,char *string);
//						char *					fgets(char *, int, File);
//						unsigned char 			fexist_(unsigned char*, unsigned char *, File *file);
//........................................................................................................................................
// ext. functions:		extern unsigned char SDC_GetSector  (unsigned long,unsigned char *);		
//						extern unsigned char SDC_PutSector (unsigned long,unsigned char *);
//........................................................................................................................................
//
// URL:					www.Mikro-Control.de
// mailto:				stephan.busker@mikro-control.de
//________________________________________________________________________________________________________________________________________




//________________________________________________________________________________________________________________________________________
// 
// Global variables needed for read- or write-acces to the FAT16- filesystem.
//				
//________________________________________________________________________________________________________________________________________

unsigned char 	SectorsPerCluster 		= 0;			// how many sectors does a cluster contain?
unsigned char		FatCopies 				= 0;			// Numbers of copies of the FAT
unsigned int		PossibleRootEntries 	= 0;			// Possible number of entries in the root directory.
unsigned int		SectorsPerFat 			= 0;			// how many sectors does a fat16 contain?
unsigned long		ReservedSectors 		= 0;			// Sectors reserved by the filesystem.
unsigned long		FirstPartitionSector 	= 0;			// Distance in sectors between the first partition and the master bootrecord.
unsigned long		FileAllocationTable 	= 0;			// pointer to the first FAT
unsigned long		RootDirectory 			= 0;			// Pointer to the rootdirectory of the first partition.
unsigned long		FirstDataCluster 		= 0;			// Pointer to the first cluster containing data (cluster0).

unsigned char   	FileBuffer[512];						// Buffer for read and write operation from or to the mmc.
struct DirEntry 	*DirectoryEntry;						// Pointer to an entry of the directory.
struct FatEntry 	*Fat;									// Pointer to an entry of the fat (next clusterposition).




//________________________________________________________________________________________________________________________________________
// Funtion: 	InitFat16(void);
// 
// Description:	This function reads the Masterbootrecord and finds the position of the Volumebootrecord, the FAT and the Rootdirectory
//				and stores the information in global variables.
//
// Returnvalue: The function returns "0" if the filesystem could not be initialized because no partition was found on the disc. 
//________________________________________________________________________________________________________________________________________

unsigned char InitFat16(void)
{	
	SDC_GetSector((unsigned long)MBR_SECTOR,&FileBuffer[0]);				// Read the master bootrecord from mmc.	
	FirstPartitionSector= *((unsigned long*)&FileBuffer[0x1C6]);				// Position of the first datacluster.	
	SDC_GetSector((unsigned long)FirstPartitionSector,&FileBuffer[0]);		// Read the volume bootrecord from mmc.	
	SectorsPerCluster	= FileBuffer[0x0D];										// Number of sectors per cluster. Depends on the memorysize of the sd-card.
	FatCopies 			= FileBuffer[0x10];										// Number of fatcopies.
	PossibleRootEntries = *((unsigned int*) &FileBuffer[0x11]);				// How many Entries are possible in the rootdirectory (FAT16 allows max. 512 entries).
	SectorsPerFat 		= *((unsigned int*) &FileBuffer[0x16]);
	ReservedSectors		= *((unsigned int*)&FileBuffer[0x0E]);				// calculate the sectorpositon of the FAT, the Rootdirectory and the first Datacluster.
	FileAllocationTable	=   (unsigned long)((unsigned long)FirstPartitionSector + (unsigned long)ReservedSectors);
	RootDirectory		=   (unsigned long)((unsigned long)FileAllocationTable + (unsigned long)((unsigned long)SectorsPerFat*(unsigned long)FatCopies));	
	FirstDataCluster	=   (unsigned long)((unsigned long)RootDirectory + ((unsigned long)(PossibleRootEntries>>4)));
	return(0);	
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned char fopen_(unsigned char*, unsigned char *, File *file);
// 
// Description:	This function looks for the specified file in the rootdirectory of the drive. If the file is found the number of the 
//				corrosponding datacluster is returned to main. Only modes 'r' (reading) and 'a' append are implemented yet.
//
//	Return:		0 = faild to open speicified file
//				1 = file opened 
//________________________________________________________________________________________________________________________________________

unsigned char fopen_(unsigned char *fname, char mode, File *file)
{
	unsigned long cluster_temp	= 0;
	unsigned char name[11] 	= "           ";			
	unsigned int temp  		= 0;
	
	if(file == 0) return(0);														// no valid pointer 

	file->start_cluster			= 0;												// Sectorpointer to the first sector of the first datacluster of the file. 
	file->cluster_pointer		= 0;												// Pointer to the cluster which is edited at the moment.
	file->sector_index 			= 0;												// The sector which is edited at the moment (cluster_pointer + sector_index).
	file->byte_index 			= 0;												// The bytelocation within the current sector (cluster_pointer + sector_index + byte_index).
	file->mode 					= mode;												// mode of fileoperation (read,write)
	file->filesize 				= 0;												// the size of the opend file in bytes.
	file->fileposition 			= 0;												// pointer to a character within the file 0 < fileposition < filesize
	file->sector_in_buffer 		= 0;												// the last sector read, wich is still in the sectorbuffer.
	file->directory_sector 		= 0;												// the sectorposition where the directoryentry has been made.
	file->directory_index 		= 0;												// the index to the directoryentry within the specified sector.
	file->attribute 			= 0;												// the attribute of the file opened.

	if(ScanSubDirectories(&fname[0], file))											// Is the specified filepath available?
	{
		SeperateFileName(&fname[0], &name[0]);											// seperate the filename and attribute from the filepath and bring them in the correct format.			
		file->attribute = _FILE;
		if(SeekDirectoryEntry(&name[0], file))											// if file was found
		{
			if(mode == 'a')																// open existing file for writing (append data at the end of the file)
			{	
				fseek_(file, 0, SEEK_END);												// fseek points to the end of the file
			}
			return(1);
		}
		else
		{
			if((mode == 'a') || (mode == 'w'))											// specified file doesn't exist so create new file for writing data.
			{	
				cluster_temp = (unsigned long)FindNextFreeCluster(file);				// the next free cluster on the disk.
				if(cluster_temp) 														// if a free cluster is available:
				{
					temp = (unsigned int)cluster_temp;									// remember the index of the free datacluster found for the directory entry.
					cluster_temp -=2;													// Clusterposition is ((position in FAT)-2). first two entries in FAT are reserved.
					cluster_temp *= SectorsPerCluster;									// Calculate relative sectorindex of first datacluster.
					file->start_cluster   = (FirstDataCluster + cluster_temp);			// Calculate absolute sectorposition of first datacluster.
					file->cluster_pointer = file->start_cluster;						// start reading the file with the first sector of the first datacluster.
					if(CreateDirectoryEntry(&name[0],temp,file,_FILE))					// Could an entry for the new file in the rootdirectory be created?
					{
						return(1);
					}	
				}				
			}
		}
	}
	return(0);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	fflush_(File *file);
// 
// Description:	This function writes the data already in the buffer but not yet written to the file. 
//				
//________________________________________________________________________________________________________________________________________

int	fflush_(File *file)
{
	unsigned int time=0;
	unsigned int date=0;

#ifdef __USE_TIME_DATE_ATTRIBUTE															// has the date and time attribute of the file to be set?
	
	time = (((rtctime.hour)<<11)     		| ((rtctime.minute)<<5) 	| rtctime.second); 
	date = ((((rtctime.year)-1980) <<9) 	| ((rtctime.month) <<5) 	| rtctime.day); 

#endif	
	
		
	if(file && file->mode =='a')
	{
		if(file->byte_index > 0)															// has data been added to the file?		
		{
			SDC_PutSector((unsigned long)(file->cluster_pointer + file->sector_index),&FileBuffer[0]);	// save the data still in the buffer 	
		}	
		SDC_GetSector((unsigned long)file->directory_sector,&FileBuffer[0]);			// read the directoryentry for this file.
		DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
		DirectoryEntry[file->directory_index].size = (unsigned long) file->filesize;
		DirectoryEntry[file->directory_index].time = (unsigned int) time;
		DirectoryEntry[file->directory_index].date = (unsigned int) date;
		SDC_PutSector((unsigned long)file->directory_sector,&FileBuffer[0]);
	}
	return(0);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	fclose_(File *file);
// 
// Description:	This function closes the open file by writing the remaining data from the buffer to the device and entering the filesize 
//				in the directory entry.
//________________________________________________________________________________________________________________________________________

void fclose_(File *file)
{
	fflush_(file);

	file->start_cluster			= 0;														// Sectorpointer to the first sector of the first datacluster of the file. 
	file->cluster_pointer		= 0;														// Pointer to the cluster which is edited at the moment.
	file->sector_index 			= 0;														// The sector which is edited at the moment (cluster_pointer + sector_index).
	file->byte_index 			= 0;														// The bytelocation within the current sector (cluster_pointer + sector_index + byte_index).
	file->mode 					= 0;														// mode of fileoperation (read,write)
	file->filesize 				= 0;														// the size of the opend file in bytes.
	file->fileposition 			= 0;														// pointer to a character within the file 0 < fileposition < filesize
	file->sector_in_buffer 		= 0;														// the last sector read, wich is still in the sectorbuffer.
	file->directory_sector 		= 0;														// the sectorposition where the directoryentry has been made.
	file->directory_index 		= 0;														// the index to the directoryentry within the specified sector.
	file->attribute 			= 0;														// the attribute of the file opened.
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	long fread_(void *buffer, long size, long count, File *file);
// 
// Description:	This function reads count objects of the specified size from the actual position of the file to the specified buffer.
//				
// Returnvalue: The function returns the number of objects (not bytes) read from the file.	
//				
//________________________________________________________________________________________________________________________________________

unsigned long fread_(void *buffer, unsigned long size, unsigned long count, File *file)
{
	unsigned long object_cnt 	= 0;														// count the number of objects read from the file.
	unsigned long object_size 	= 0;														// count the number of bytes read from the actual object.
	unsigned char *buff_pnt    = 0;														// a pointer to the actual bufferposition.
	unsigned char success      = 1;														// no error occured during read operation to the file.
	
	buff_pnt = (unsigned char *) buffer;													// cast the void pointer to an unsigned char *
	
	while((object_cnt < count) && success)
	{
		object_size = size;
		while((size > 0) && success)
		{
			*buff_pnt = (unsigned char) fgetchar_(file);									// read a byte from the buffer to the opened file.
			buff_pnt++;
			size--;
		}
		if(success) object_cnt++;
	}		
	
	return(object_cnt);																	// return the number of objects succesfully read from the file
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	long fwrite_(void *buffer, long size, long count, File *file);
// 
// Description:	This function writes count objects of the specified size from the buffer to the actual positon within the file.
//				
// Returnvalue: The function returns the number of objects (not bytes) written to the file.	
//				
//________________________________________________________________________________________________________________________________________

unsigned long fwrite_(void *buffer, unsigned long size, unsigned long count, File *file)
{
	unsigned long object_cnt 	= 0;														// count the number of objects written to the file.
	unsigned long object_size 	= 0;														// count the number of bytes written from the actual object.
	unsigned char *buff_pnt    = 0;														// a pointer to the actual bufferposition.
	unsigned char success      = 1;														// no error occured during write operation to the file.
	
	buff_pnt = (unsigned char *) buffer;													// cast the void pointer to an unsigned char *
	
	while((object_cnt < count) && success)
	{
		object_size = size;
		while((size > 0) && success)
		{
			success = fputchar_(file, *buff_pnt);											// write a byte from the buffer to the opened file.
			buff_pnt++;
			size--;
		}
		if(success) object_cnt++;
	}		
	
	return(object_cnt);																	// return the number of objects succesfully written to the file
}																							// (!!!!! objects and not bytes !!!!)

//________________________________________________________________________________________________________________________________________
// Funtion: 	int fseek_(File *, long, int)
// 
// Description:	This function sets the pointer of the stream relative to the position 
//				specified by origin (SEEK_SET, SEEK_CUR, SEEK_END). 
//				
//________________________________________________________________________________________________________________________________________

int fseek_(File *file, long offset, int origin)
{
	long			fposition 	= 0;
	int 			retvalue 	= 1;
	unsigned long 	temp 		= 0;
	
//......................................................
	if(origin == SEEK_SET)												// Fileposition relative to the beginning of the file.
	{
		fposition = 0;
	}	
//......................................................
	else if(origin == SEEK_END)										// Fileposition relative to the end of the file.
	{
		fposition  = (long) file->filesize;
	}	
//......................................................
	else if(origin == SEEK_CUR)										// Fileposition relative to the current position of the file.
	{
		fposition = file->fileposition;
	}	

	fposition += offset;

	if((fposition >= 0) && (fposition <= (long)file->filesize))		// is the pointer still within the file?
	{
		retvalue 				= 0;
		file->sector_index 		= 0;
		file->byte_index   		= 0;
		file->fileposition 		= 0;
		file->cluster_pointer 	= file->start_cluster;
		
		while(file->fileposition < fposition)
		{
			file->fileposition++;										
			if(file->byte_index < 511)									
			{
				file->byte_index++;
			}
			else
			{
				file->byte_index=0;										// reading at the beginning of new sector.
				file->sector_index++;									// continue reading in next sector
				if(file->sector_index >= SectorsPerCluster)			// When end of cluster is reached, the next datacluster has to be searched in the FAT.
				{
					if(file->fileposition < fposition) 
					{
						file->sector_index = 0;							// start reading new cluster at first sector of the cluster.
						GetNextCluster(file);							// Sets the clusterpointer of the file to the next datacluster.
					}
				}
			}	
		}
		if(file->byte_index)				
		{
			temp = (unsigned long)((unsigned long)file->cluster_pointer + (unsigned long)file->sector_index);			
			SDC_GetSector((unsigned long)temp,&FileBuffer[0]);		// FileBuffer will be written at once at the end of the cluster and has to be updated.						
		}
	}	
	return(retvalue);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	fgetchar_(File *file);
// 
// Description:	This function reads and returns one character from the specified file. Is the end of the actual sector reached the
//				next sector of the cluster is read. Is the last sector of the cluster read the next cluster will be searched in FAT.
// Returnvalue: The function returns the character read from the specified memorylocation as unsigned char casted to int or EOF.
//________________________________________________________________________________________________________________________________________

int fgetchar_(File *file)
{	
	int c = EOF;
	unsigned long temp1;
	
	if(file->filesize > 0)														// wen the end of the file is not reached, get the next character.
	{
		temp1  = (unsigned long)file->cluster_pointer;						// calculate the adress of the next character to be read.
		temp1 += (unsigned long)file->sector_index;			
	
		if(file->sector_in_buffer != temp1)									// Has the content of the buffer been modified and has to be updated?
		{
			SDC_GetSector((unsigned long)temp1,FileBuffer);					// Read the calculated cluster.			
			file->sector_in_buffer = (unsigned long)temp1;
		}	
		c = (int) FileBuffer[file->byte_index];
		file->filesize--;														// decrement the number of characters available.
	
		if(file->byte_index < 511)												// continue reading from this sector until the end of the sector is reached.
		{
			file->byte_index++;
		}
		else																	// has the end of an sector been reached->
		{
			file->byte_index=0;													// continue reading at the beginning -
			file->sector_index++;												// of new sector.
			if(file->sector_index >= SectorsPerCluster)						// When the end of an cluster is reached, the next datacluster has to be searched in the FAT.
			{
				file->sector_index = 0;											// start reading new cluster at first sector of the cluster.
				GetNextCluster(file);											// Sets the clusterpointer of the file to the next datacluster.
			}
		}
	}
	return(c);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	fputchar_(File *file);
// 
// Description:	This function writes a byte to the specified file and takes care of writing the necessary FAT- Entries. 
//				
// Returnvalue: The function returns a value of 0 if the data could not be written.
//________________________________________________________________________________________________________________________________________

unsigned char fputchar_(File *file,char c)
{	
	unsigned long ul_temp  = 0;
	unsigned char retvalue = 1;
	
    if(file->sector_index >= SectorsPerCluster)										// if end of the cluster is reached, find next free cluster
	{
		file->sector_index = 0;
		if(!AppendCluster(file)) retvalue = 0;											// append a new and free cluster at the end of the file.
	}

	FileBuffer[file->byte_index] = c;													// write databyte into the buffer. The byte will be written to the device at once
	if(file->filesize == file->fileposition) file->filesize++;							// a character has been written to the file so the size is inkremented but only when the character has been added at the end of the file.
	file->fileposition++;																// the actual positon within the file. 
																						// if the buffer contains the complete sectordata.
	if(file->byte_index < 511)															// if the end of this sector is not reached yet
	{
		file->byte_index++;																// the next byte will be written to the next byteposition in this sector.
	}
	else																				// otherwise the data in the sectorbuffer will be written to the device and the next sector will be selected.
	{
		ul_temp  = (unsigned long)file->cluster_pointer;
		ul_temp += (unsigned long)file->sector_index;
		
		SDC_PutSector((unsigned long)ul_temp,&FileBuffer[0]);
		file->byte_index=0;																// and the next byte will be written at the beginning of this new sector.
		file->sector_index++;
		if(file->sector_index >= SectorsPerCluster)									// if end of the cluster is reached, find next free cluster
		{
			file->sector_index = 0;
			if(!AppendCluster(file)) retvalue = 0;										// append a new and free cluster at the end of the file.
		}
	}
	return(retvalue);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	fputs_(File *file);
// 
// Description:	This function writes a string to the specified file.
//				
//________________________________________________________________________________________________________________________________________

unsigned char fputs_(File *file,char * string)
{
	unsigned char i=0;
	
	while(string[i] != 0)
	{
		fputchar_(file,string[i]);
		i++;
	}
	return(0);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	fgets_(File *file);
// 
// Description:	This function reads a string from the file to the specifies string.
//		
// Returnvalue: A pointer to the string written from the file.
//________________________________________________________________________________________________________________________________________

char * fgets_(char *string, int count, File *file)
{
	int buff_pnt = 0;
	int buff_tmp = 0;
	unsigned char state = 0;															// Variable used for a statemachine to recognize the end of a line.
	
	while(count > 1)																	// read the count-1 characters from the file to the string.
	{
		buff_tmp = fgetchar_(file);														// read a character from the opened file.
		switch(state)
		{
			case 0:
				if(buff_tmp == 0x0D) state++;											
			break;
			
			case 1:
				if(buff_tmp == 0x0A)
				{
					count = 1;
				}
				state = 0;	
			break;
		}
		if(buff_tmp == EOF) {buff_tmp = 0; count = 1;}									// is the end of the file reached, terminate the string with zero.
		string[buff_pnt] = (char) buff_tmp;
		count--;
		buff_pnt++;
	}
	string[buff_pnt] = 0;
	return(string);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned char fexist_(unsigned char*, unsigned char *, File *file);
// 
// Description:	This function searches the specified file and returns 0 if the file was not found.
//				
//
//	Return:		0 = file does not exist
//				1 = file exists
//________________________________________________________________________________________________________________________________________

unsigned char fexist_(unsigned char *fname, File *file)
{
	if(fopen_(fname,'r', file))
	{
		fclose_(file);
		return(1);
	}
	else
	{
		return(0);
	}
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	GetNextCluster(File *file);
// 
// Description:	This function finds the next datacluster of the file specified with File *file. 
//				
// Returnvalue: The function returns "0" if the last cluster has already been reached.
//________________________________________________________________________________________________________________________________________

unsigned int GetNextCluster(File *file)
{
	unsigned long fat_pointer			=	0;
	unsigned long fat_sector_offset	=	0;
	unsigned long ul_tmp				=  	0;
	unsigned char retvalue				= 	0;										// no new cluster found yet.
	
	
	if((file->cluster_pointer >= RootDirectory) && (file->cluster_pointer < (RootDirectory + 31)))
	{																				// Is the next cluster searched within the rootdirectory and available?
		file->cluster_pointer++;													// the rootdirectory is a linear adress space of 32 clusters.
		retvalue = 1;																// and the next cluster has been found.
	}
	else if(file->cluster_pointer > (RootDirectory + 31))							// The specified cluster is within the FAT.
	{
		fat_sector_offset   = ((file->cluster_pointer) - (FirstDataCluster));		// Calculate index of actual cluster within the FAT.
		fat_sector_offset  /= SectorsPerCluster;									// calculate the index of the actual sector within the FAT. 					
		fat_sector_offset  += 2;													// In Fat16 clusterpositions have an offset of two.	
		fat_pointer = (fat_sector_offset%0x100);									// Calculate the sector within the cluster.																
		fat_sector_offset   = (fat_sector_offset>>8);								// and the position within the sector.

		SDC_GetSector((unsigned long)(FileAllocationTable + fat_sector_offset),&FileBuffer[0]);	
		file->sector_in_buffer = (FileAllocationTable + fat_sector_offset);			// Mark that new sector has been read.
	
		ul_tmp  = (unsigned long)FileBuffer[((fat_pointer << 1)+1)];				// Read next sector information from calculated clusterposition.
		ul_tmp  = (ul_tmp << 8);
		ul_tmp |= (unsigned long)FileBuffer[(fat_pointer << 1)];    
		ul_tmp -=2;																	// next datacluster is clusterposition in fat - 2.
		ul_tmp *= SectorsPerCluster;												// calculate sectorposition of new cluster
		ul_tmp += FirstDataCluster;													// in relation to first datacluster of the disk.
	
		if(ul_tmp < 0xfff7)															// has a new cluster been read or was the end of the fat reached?
		{
			file->cluster_pointer = (unsigned long) ul_tmp;						// continue reading the file at the beginning of new datacluster.
			retvalue = 1;															// a new cluster was found.
		}	
	}
	return(retvalue);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int FindNextFreeCluster(void)
// 
// Description:	This function looks in the FAT to find the next free datacluster
//				
// Returnvalue: The function returns the adress of the next free cluster found within the fAT.
//________________________________________________________________________________________________________________________________________

unsigned int FindNextFreeCluster(File *file)
{
	unsigned long fat_pointer			=	0;							// Pointer to the first sector of the FAT.
	unsigned long ul_tmp				=  	0;							// temporary variable used to calculate a sectornumber.
	unsigned int  fat_sector_offset	=	0;							// index to a sector within the FAT.
	unsigned int  fat_entry			= 	0;							// index to an fatentry within the actual sector (256 fatentries are possible within one sector).
	unsigned int  free_cluster			= 	0;							// a pointer to the first sector of the next free cluster.
	
	fat_pointer = (unsigned long) FileAllocationTable;					// start searching for empty cluster at the beginning of the fat.
																		// if the end of the fat is not reached yet and no free cluster has been found
	while((fat_sector_offset < SectorsPerFat) && (!free_cluster))	
	{
		ul_tmp = (unsigned long) ((unsigned long)fat_pointer + (unsigned long)fat_sector_offset);
		SDC_GetSector((unsigned long)ul_tmp,&FileBuffer[0]);				// read next sector of FAT.
		file->sector_in_buffer = ul_tmp;								// remember the number of the sector in FileBuffer.
		Fat = (struct FatEntry *)&FileBuffer[0];
		for(fat_entry=0;fat_entry<256;fat_entry++)						// look for an free cluster at all entries in this sector of the fat. 
		{
			if(Fat[fat_entry].next_cluster == 0x0000)					// empty cluster found!!
			{				
				Fat[fat_entry].next_cluster = 0xffff;					// mark this fat-entry as used and save it to the device.
				SDC_PutSector((unsigned long)file->sector_in_buffer,&FileBuffer[0]);
				free_cluster  = fat_entry;								// the relative position of the free cluster found in this sector of the FAT.
				free_cluster += (fat_sector_offset << 8);				// calculate the absolute position of the free cluster in the FAT;
				fat_entry = 256;										// terminate the search for a free cluster in this sector.
			}
		}
		fat_sector_offset++;									
	}	
return(free_cluster);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int AppendCluster(File *file);
// 
// Description:	This function finds the next free datacluster on the disk and appends it to the specified file.
//				
// Returnvalue: This funktion returns 1 if a cluster was appended to the specified file.
//________________________________________________________________________________________________________________________________________

unsigned char AppendCluster(File *file)
{
	unsigned int  free_cluster = 0;														
	unsigned long fat_pointer	= 0;
	unsigned char retvalue    	= 0;
	
	free_cluster = FindNextFreeCluster(file);												// the next free cluster found on the disk.
	if(free_cluster) retvalue = 1;															// A free cluster was found and can be added to the end of the file. 
	fat_pointer  = FileAllocationTable;														// Set Pointer to the beginnig of the FAT.
	fat_pointer += (unsigned long)((unsigned long)GetFatClusterOffset(file) >> 8);		// find the sector in the FAT with 256 entries per sector.
	
    SDC_GetSector(fat_pointer,&FileBuffer[0]);                                            
	Fat = (struct FatEntry *)&FileBuffer[0];
	Fat[GetFatSectorIndex(file)].next_cluster = free_cluster;								// append the free cluster to the end of the file in the FAT.
	SDC_PutSector((unsigned long)fat_pointer,&FileBuffer[0]);							// save the modified sector to the FAT.

	fat_pointer  = (unsigned long)free_cluster;
	fat_pointer -= 2;
	fat_pointer *= SectorsPerCluster;
	fat_pointer += FirstDataCluster;

	file->cluster_pointer = fat_pointer;													// continue wrtiting to the file in the new and free datacluster.
	return(retvalue);																		// return 1 when a new cluster was appended to the file
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int GetFatClusterIndex(File *file);
// 
// Description:	This function returns the clusterindex of the cluster specified by file->cluster_pointer of the specified file.
//				
//________________________________________________________________________________________________________________________________________

unsigned int GetFatClusterOffset(File *file)
{
	unsigned long fat_sector_offset	=	0;

	fat_sector_offset   = ((file->cluster_pointer) - (FirstDataCluster));				// Calculate index of actual cluster in FAT.
	fat_sector_offset  /= SectorsPerCluster;														
	fat_sector_offset  += 2;															// In Fat16 clusterpositions have an offset of two.								
	
	return((unsigned int)fat_sector_offset);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int GetFatSectorIndex(File *file);
// 
// Description:	This function returns the sectorindex of the cluster specified by file->cluster_pointer of the specified file.
//				
//________________________________________________________________________________________________________________________________________

unsigned int GetFatSectorIndex(File *file)
{
	unsigned int fat_pointer			=	0;
	
	fat_pointer = GetFatClusterOffset(file);
	fat_pointer = fat_pointer % 0x100;													// Calculate the clusterposition in the fat																
	
	return(fat_pointer);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int CreateDirectoryEntry(unsigned char *, unsigned int, unsigned char attrib)
// 
// Description:	This function looks for the next free position in the directory and creates an entry. The type of an directoryentry is
//				specified by the attribute. 
//				bit0:	unused
//				bit1:	archive
//				bit2:	read_only
//				bit3:	system  
//				bit4:	directory
//				bit5:	volume
//________________________________________________________________________________________________________________________________________

unsigned char CreateDirectoryEntry(unsigned char *fname, unsigned int cluster, File *file,unsigned char attrib)
{
	unsigned int 	rootentry	= 0;													// index to an entry in the rootdirectory.
	unsigned int 	cnt_enries_searched = 0;											// count the number of rootentries which have been searched already.
	unsigned char 	i = 0;
	unsigned int 	sector_offset = 0;													// index to the sector of the Rootentry which is searched momentarily
	unsigned char 	retvalue = 0;
	
																						// directory starts at sector specified by dir_sector. This can be the rootdirectory or any other directory.
	do
	{																					// search the next 16 rootentries in this sector of the roordirectory.
		rootentry=0;									
		SDC_GetSector((unsigned long)(RootDirectory + sector_offset),&FileBuffer[0]);	// Read the Rootdirectory.
		DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
		while((rootentry<16) && (!retvalue))
		{
			if((DirectoryEntry[rootentry].attribute == 0) || (DirectoryEntry[rootentry].attribute == 0xE5))	// empty directory entry found
			{
				for(i=0;i<11;i++) DirectoryEntry[rootentry].name[i] = fname[i];		// Kopie the filename and the file extension to the directoryentry.
				DirectoryEntry[rootentry].attribute    = attrib;						// Set the fileattribute to archive to reserve the directoryentry.
				DirectoryEntry[rootentry].startcluster = cluster;						// copy the location of the first datacluster to the directoryentry.
				DirectoryEntry[rootentry].size     = 0;									// the new createted file has no content yet.
				file->directory_sector = (unsigned long) (RootDirectory + sector_offset);
				file->directory_index  = (unsigned char) rootentry;
				retvalue = 1;
				SDC_PutSector((unsigned long)(RootDirectory + sector_offset),&FileBuffer[0]);				
			}			
			rootentry++;
			cnt_enries_searched++;
		}
		if(!retvalue)																	// file not found in this sector so take next sector.
		{
			rootentry = 0;
			sector_offset++;
		}
	}
	while((cnt_enries_searched< PossibleRootEntries) && (!retvalue));

	return(retvalue);																	// return 1 if file has been created otherwise return 0.
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	unsigned int SeekDirectoryEntry(unsigned char*, File *, unsigned char)
// 
// Description:	this function searches all possible rootentries until the file or directory is found or the end of the rootdirectory is reached
//				
// Returnvalue: This function returns 1 if the directoryentry specified was found.
//________________________________________________________________________________________________________________________________________

unsigned char SeekDirectoryEntry(unsigned char *fname, File *file)
{
	unsigned int 	rootentry=0;
	unsigned int   end_of_directory_not_reached = 0;		// the directory has been read completely without a result.
	unsigned char 	i=0;
	unsigned char  retvalue = 0;
	unsigned long  cluster_temp = 0;
	unsigned int cnt = 0;
															// directory starts at sector specified by dir_sector. This can be the rootdirectory or any other directory.
	do
	{														// search the next 16 rootentries in this sector of the roordirectory.
		rootentry=0;									
		SDC_GetSector((unsigned long) file->cluster_pointer,&FileBuffer[0]);		// Read the Rootdirectory.
		DirectoryEntry = (struct DirEntry *)&FileBuffer[0];
		cnt++;
		while((!retvalue)&&(rootentry<16))
		{
			i=0;
			while((i<10)&&(DirectoryEntry[rootentry].name[i] == fname[i]))
			{
				i++;
			}
			if((i==10)  && (DirectoryEntry[rootentry].attribute == file->attribute))		// entry found!! -> reading startcluster of entry from offset 26.
			{
				cluster_temp = (unsigned long)DirectoryEntry[rootentry].startcluster; 							
				cluster_temp -=2;															// Clusterposition is ((position in FAT)-2). first two entries in FAT are reserved.
				cluster_temp *= (unsigned long)SectorsPerCluster;							// Calculate positon of first cluster.
				file->start_cluster   = (FirstDataCluster + cluster_temp);
				file->directory_sector = (unsigned long) file->cluster_pointer;
				file->cluster_pointer = file->start_cluster;								// start reading the file with the first cluster.
				file->filesize = (unsigned long) DirectoryEntry[rootentry].size;
				file->directory_index  = (unsigned char) rootentry;
				retvalue = 1;
			}
			rootentry++;
		}
		if(!retvalue)											// file not found in this sector so take next sector.
		{
			end_of_directory_not_reached = GetNextCluster(file);
		}
	}
	while((end_of_directory_not_reached) && (!retvalue));
	return(retvalue);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	void ScanSubDirectories(unsigned char*, File *file);
// 
// Description:	This function scans the filename for subdirectories and changes the directory until the last directory is reached. 
// 				here the specified file is searched. 
//				
// Returnvalue:
//________________________________________________________________________________________________________________________________________

unsigned char ScanSubDirectories(unsigned char *fname, File *file)
{
	unsigned char readpointer	= 0;						// pointer to read a character within the filepath.
	unsigned char writepointer	= 0;						// pointer to write a character into the string dirname.
	unsigned char dircnt 		= 0;						// the number of subdirectories found in filepath.
	unsigned char dirname[11];								// temporary variable containing the name of a directory within the filepath.
	unsigned char retvalue     = 1;						// no error opening a subdirectory occured yet.	
	unsigned char cnt			= 0;						// maximun number of characters in a path is 256;
	

	while((fname[readpointer]!=0) &&  (cnt < 255))		// search the string until the end is reached.	
	{
		cnt++;
		if((fname[readpointer] == '/') && readpointer)		// subdirectory found. ignore first "/" as directory. 
		{
			dircnt++;										// count the number of subdirectories found within the filepath.		
		}
		readpointer++;
	}
	
	for(writepointer=0;writepointer<10;writepointer++) dirname[writepointer] = ' ';
	writepointer = 0;
	readpointer = 0;										// start scanning the string at the beginning.

	file->cluster_pointer = RootDirectory;					// always start searching in the rootdirectory.
	file->attribute = _DIRECTORY;							// the attribute of the item to be searched in the specified directory.	
	
	while(dircnt && retvalue)								// scan all subdirectories found before.
	{
		if(fname[readpointer] != '/')						// is the end of the subdirectory entry not reached yet?
		{
			if((fname[readpointer]>96) && (fname[readpointer]<123))
			{												// all characters must be upper case.
				dirname[writepointer]=(fname[readpointer] - 32);					
			}
			else
			{
				dirname[writepointer]=fname[readpointer];
			}	
			writepointer++;
		}	
		else												
		{			
			dircnt--;
			if(!(SeekDirectoryEntry(dirname, file)))		// was the seperated subdirectory not found?
			{
				retvalue = 0;								// leave the function with return(0) otherwise continue with the next subdirectory until all directories have been searched.
			}
			for(writepointer=0;writepointer<10;writepointer++) dirname[writepointer] = ' ';
			writepointer = 0;
		}
		readpointer++;
	}
	file->attribute = _FILE;								// All subdirectories have been searched. the next thing to find is the specified file.
	return(retvalue);
}

//________________________________________________________________________________________________________________________________________
// Funtion: 	void SeperateFileName(unsigned char*);
// 
// Description:	This function seperates the filename and the fileattribute and brings them into the needed format ('test.txt' -> 'TEST    TXT');
//				
//________________________________________________________________________________________________________________________________________

void SeperateFileName(unsigned char *fname, unsigned char *name)
{
	unsigned char readpointer	= 0;
	unsigned char writepointer	= 0;
	unsigned char dircnt 		= 0;						// the number of subdirectories found in filepath.
	unsigned char fileindex	= 0;						// the position in the filenamestring where the last "/" is found.
	
	while(fname[readpointer]!=0)							// search the string until the end is reached.	
	{
		if((fname[readpointer] == '/') && readpointer)		// subdirectory found. ignore first "/" as directory. 
		{
			dircnt++;										// subdirectory entry found.		
			fileindex = (readpointer + 1);						// store the index of the last slash found as the beginning of the filename.
		}
		readpointer++;
	}
	readpointer = fileindex;								// start seperating the filename from the directory at the stored position.
	dircnt = 0;

	while((writepointer<=10) && (fname[readpointer]!=0))	// the rootdirectoryentry is 8bytes for filename and 3bytes for fileattribute.
	{														// the filename in the rootdirectory is in the format "TEST    TXT" without the dot.
		if(fname[readpointer]=='.')						// seperating filename and attribute.
		{
			readpointer++;
			writepointer=8;			
		}
		else
		{
			if((fname[readpointer]>96) && (fname[readpointer]<123))
			{
				name[writepointer]=(fname[readpointer] - 32);					// all characters must be upper case.
			}
			else
			{
				name[writepointer]=fname[readpointer];
			}	
			readpointer++;
			writepointer++;
		}
	}	
}
